/*
 * vifctl/vif.h
 * $Rev: 1259 $
 *
 * Copyright (C) 2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#ifndef __VIF_H
#define __VIF_H

/**
 * @file vif.h
 * @brief Interface to the virtual wireless device implementation.
 *
 * It allows to directly query any wireless parameter of a virtual wireless
 * interface or access it via wireless ioctl() calls.
 *
 * The shared memory segment is transparently created when invoking vif_create()
 * and attached for read-write. A client can attach it read-only with
 * vif_get_names(). Calling vif_get_names() after vif_create() doesn't change
 * the read-write status.
 *
 * 1) Virtual interface standard ioctls:
 * If you want to add another ioctl() type, add it to the ioctl_handler_list at
 * the end of the file with the corresponding function calls vif_siocgiw...()
 * for getter and vif_siocsiw...() for setter function.
 *
 * 2) Virtual interface private ioctls:
 * If you want to add a private ioctl() call, extend the vif_private_args below.
 *
 * 3) Virtual interface driver capabilities:
 * If you want to change the default interface capabilities, change the values
 * under "DEFAULT VIF PARAMETERS" below or in vif_config_init_iwrange().
 *
 * 4) Initial virtual interface configuration:
 * If you want to change the initial interface state, change the parameter
 * assignment in vif_config_init_wlan().
 */

/* INCLUDES *******************************************************************/

#include <netinet/in.h>             // struct in_addr (for wireless.h)
#include <linux/wireless.h>         // iwreq, SIOCGIWNAME, SIOCGIWNWID, ...
#include "iw_compat.h"              // wireless_info



/* CONSTANTS DEFINITIONS ******************************************************/

/** Maximum number of virtual wireless interfaces */
#define MAX_NUM_VIF     2



/* LOCAL DATA TYPES ***********************************************************/

/**
 * @struct vif_info
 * Virtual wireless interface information
 */
typedef struct vif_info
{
    char vif_name[IFNAMSIZ + 1];        /**< Virtual interface name */
    unsigned char vif_mac[ETH_ALEN];    /**< Virtual interface Ethernet address */
    struct wireless_info vif_wlaninfo;  /**< Virtual interface wireless configuration */
}
vif_info;

/**
 * Assign ioctl() type to corresponding function handling the ioctl()'s iwreq
 * structure
 */
typedef struct ioctl_handler
{
    int ioctl_num;                      /**< Ioctl() type */
    int (*iw_handler)(struct iwreq * request);  /**< Function pointer to request handler */
}
ioctl_handler;

/* Config access mode */
#define CONFIG_READONLY     0
#define CONFIG_READWRITE    1



/* DEFAULT VIF PARAMETERS (according to iwrange of ath_pci) *******************/

/** Frequency of each channel in MHz */
static const long
channel_frequency[] = {
    2412, 2417, 2422, 2427, 2432, 2437, 2442,
    2447, 2452, 2457, 2462, 2467, 2472
};

/* Driver parameters for newly created interfaces */
#define VIF_WE_VERSION      22      /**< Wireless extension version compatibility */
#define VIF_NUM_FREQ        sizeof(channel_frequency) / sizeof(long)
                                    /**< Number of frequencies */
#define VIF_NUM_CHANNEL     VIF_NUM_FREQ /**< Number of channels */
#define VIF_MAX_RETRY       255     /**< Max value for number of retries */
#define VIF_MAX_RTS         2347    /**< Max value for RTS threshold */
#define VIF_DEFAULT_TXPWR   17      /**< Default sending power [dBm] */

/** List of ioctls returned via SIOCGIWPRIV (for iwpriv compatibility) */
static const struct iw_priv_args vif_private_args[] = {
    { SIOCSIWMODE, IW_PRIV_TYPE_INT | IW_PRIV_SIZE_FIXED | 1, 0, "mode" }
};

/** Number of private ioctls (for convenience) */
#define VIF_IW_NUM_IOCTL_PRIV   (sizeof(vif_private_args) / sizeof(struct iw_priv_args))



/* FUNCTIONS ******************************************************************/

/** Load virtual interface configuration and return number of configured
 *  interfaces */
ssize_t
vif_init();

/** Create virtual interface (real TUN device and wireless information). */
int
vif_create(const char *);

/** Delete virtual interface (remove TUN interface and shared memory interface
 *  configuration */
int
vif_delete(const char *dev_name);

/**
 * Perform ioctl on virtual interface
 * @param request Type of ioctl() call (see wireless.h)
 * @param pwrq Wireless ioctl() request
 * @return 0 if successful, IWERR_* if error
 * @see iw_compat.h
 */
int
vif_ioctl(int request, struct iwreq *pwrq);

/** Get names and number of virtual interfaces. */
int
vif_get_names(char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1]);

/** Print virtual interface configuration list to debug output */
void
vif_print(void);

/** Return the Ethernet address (MAC) of a virtual interface */
int
vif_get_mac(unsigned char *, const char *);

/** Get wireless channel number of a virtual interface */
uint16_t
vif_get_channel(const char *);

/** Get wireless frequency of a virtual interface in Hz */
double
vif_get_frequency(const char *);

/** Get wireless sensitivity of a virtual interface in dB */
int16_t
vif_get_sensitivity(const char *);

/** Get RTS threshold of a virtual interface in Bytes */
int16_t
vif_get_rts(const char *);

/** Get fragmentation threshold of a virtual interface in Bytes */
int16_t
vif_get_frag(const char *);

/** Get radio txpower of a virtual interface in mW */
int16_t
vif_get_txpower(const char *);

/** Get number of retry attempts of a virtual interface */
int16_t
vif_get_retry(const char *);

/** Set debug output for custom redirection (default: stdout) */
void
vif_set_debug_function(int (*debug)(const char *format, ...));

/** Set error output for custom redirection (default: stdout) */
void
vif_set_error_function(int (*error)(const char *format, ...));



/* PRIVATE FUNCTIONS (DON'T CALL FROM OUTSIDE) ********************************/

int vif_siocgiwname(struct iwreq *);
int vif_siocgiwnid(struct iwreq *);
int vif_siocsiwfreq(struct iwreq *);
int vif_siocgiwfreq(struct iwreq *);
int vif_siocsiwmode(struct iwreq *);
int vif_siocgiwmode(struct iwreq *);
int vif_siocsiwsens(struct iwreq *);
int vif_siocgiwsens(struct iwreq *);
int vif_siocgiwrange(struct iwreq *);
int vif_siocgiwpriv(struct iwreq *);
int vif_siocgiwap(struct iwreq *);
int vif_siocsiwessid(struct iwreq *);
int vif_siocgiwessid(struct iwreq *);
int vif_siocgiwnickn(struct iwreq *);
int vif_siocsiwrate(struct iwreq *);
int vif_siocgiwrate(struct iwreq *);
int vif_siocsiwrts(struct iwreq *);
int vif_siocgiwrts(struct iwreq *);
int vif_siocsiwfrag(struct iwreq *);
int vif_siocgiwfrag(struct iwreq *);
int vif_siocsiwtxpow(struct iwreq *);
int vif_siocgiwtxpow(struct iwreq *);
int vif_siocsiwretry(struct iwreq *);
int vif_siocgiwretry(struct iwreq *);
int vif_siocsiwencode(struct iwreq *);
int vif_siocgiwencode(struct iwreq *);
int vif_siocgiwpower(struct iwreq *);



/* IMPLEMENTED IOCTLS *********************************************************/

/** List of supported standard ioctls (see wireless.h for details) */
static const ioctl_handler ioctl_handler_list[] = {
    { SIOCGIWNAME, vif_siocgiwname },
    { SIOCGIWNWID, vif_siocgiwnid },
    { SIOCSIWFREQ, vif_siocsiwfreq },
    { SIOCGIWFREQ, vif_siocgiwfreq },
    { SIOCSIWMODE, vif_siocsiwmode },
    { SIOCGIWMODE, vif_siocgiwmode },
    { SIOCSIWSENS, vif_siocsiwsens },
    { SIOCGIWSENS, vif_siocgiwsens },
    { SIOCGIWRANGE, vif_siocgiwrange },
    { SIOCGIWPRIV, vif_siocgiwpriv },
    { SIOCGIWAP, vif_siocgiwap },
    { SIOCSIWESSID, vif_siocsiwessid },
    { SIOCGIWESSID, vif_siocgiwessid },
    { SIOCGIWNICKN, vif_siocgiwnickn },
    { SIOCSIWRATE, vif_siocsiwrate },
    { SIOCGIWRATE, vif_siocgiwrate },
    { SIOCSIWRTS, vif_siocsiwrts },
    { SIOCGIWRTS, vif_siocgiwrts },
    { SIOCSIWFRAG, vif_siocsiwfrag },
    { SIOCGIWFRAG, vif_siocgiwfrag },
    { SIOCSIWTXPOW, vif_siocsiwtxpow },
    { SIOCGIWTXPOW, vif_siocgiwtxpow },
    { SIOCSIWRETRY, vif_siocsiwretry },
    { SIOCGIWRETRY, vif_siocgiwretry },
    { SIOCSIWENCODE, vif_siocsiwencode },
    { SIOCGIWENCODE, vif_siocgiwencode },
    { SIOCGIWPOWER, vif_siocgiwpower }
};

/** Number of available regular ioctls */
#define VIF_NUM_IOCTL   sizeof(ioctl_handler_list) / sizeof(ioctl_handler)

#endif // __VIF_H
