/*
 * iwconnect/iwc_model.c
 * $Rev: 1222 $
 *
 * Copyright (C) 2009-2010 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * @file iwc_model.c
 * @brief Implements functions for communication with the WlanModel
 */

/* INCLUDES *******************************************************************/

#include <arpa/inet.h>
#include <errno.h>                  // errno
#include <signal.h>                 // signal(), SIGALRM
//#include <stdio.h>  //TEMP!!!
#include <stdlib.h>                 // calloc()
#include <string.h>                 // strncpy(), memmove(), strerror(), ...
#include <unistd.h>                 // alarm()
#include <sys/socket.h>             // sendto(), recv()
#include <iw_compat.h>              // SIOCSIWFREQ, SIOCSIWSENS, ...
#include <vif.h>                    // vif_get_rts(), vif_get_frequency(), ...
#include "iwconnect.h"              // client_info
#include "iwc_print.h"               // iwc_debug(), iwc_error(), iwc_log()
#include "message.h"                // config_msg, WLAN_MSG_SIZE


/* GLOBAL VARIABLES ***********************************************************/

extern iwc_client_info client_info;
extern iwc_model_info model_info;
extern int net_socket;


/* REGISTRATION TIMEOUT HANDLER ***********************************************/

/** Wait until we receive ack from model */
int wait = 1;

/** Clear wait status (callback for timeout signal) */
void
clear_wait(void)
{
    wait = 0;
}


/* FUNCTIONS ******************************************************************/

/**
 *  Assemble virtual device and wireless data to register interfaces
 *  @param vif_index Virtual device index
 *  @param vif_data Virtual device registration information
 *  @param dev_name Virtual wireless device name
 */
void
assemble_registration_data(int vif_index, vif_registration *vif_data,
                            char *dev_name)
{
    unsigned char mac[ETH_ALEN];
    vif_get_mac(mac, dev_name);

    // copy interface configuration
    strncpy(vif_data->vif_name, dev_name, IFNAMSIZ + 1);
    memmove(vif_data->vif_mac, mac, ETH_ALEN);
    vif_data->vif_index = vif_index;

    // copy wireless configuration (source: struct iwparam)
    // break down signed 32bit int to signed 16bit int
    vif_data->retry = vif_get_retry(dev_name);      // max value is 255  //TODO: use 8bit?
    vif_data->rts = vif_get_rts(dev_name);          // max value is 2347
    vif_data->frag = vif_get_frag(dev_name);        // max value is 2346
    vif_data->sensitivity = vif_get_sensitivity(dev_name);
    vif_data->txpower = vif_get_txpower(dev_name);
    vif_data->channel = vif_get_channel(dev_name);
}

/**
 *  Register the host at the WlanModel
 *  @return 1 if success, -1 else
 */
int
iwc_model_register(void)
{
    int i = 0;
    int num_vif = 0;
    uint16_t message_id = 0;
    int sent_bytes = -1;
    int recv_bytes = -1;
    uint8_t success = 0;
    struct msg_header *header;
    struct registration_info *reg_info;
    char vif_names[MAX_NUM_VIF][IFNAMSIZ + 1];

    // initialize member to zero initialize struct
    struct registration_msg reg_msg = { .info.msg_id = 0 };

    // make sure we have fresh memory
    char *ack_buffer = calloc(DATA_MSG_SIZE, 1);
    memset(&vif_names, 0, sizeof(vif_names));

    header = &(reg_msg.header);
    reg_info = &(reg_msg.info);

    // set host values
    header->type = REGISTRATION_MSG;
    strncpy(header->id, client_info.host_id, 2 * ETH_ALEN + 1);
    strncpy(reg_info->name, client_info.host_info.hostname, HOST_NAME_MAX);
    memmove(&reg_info->addr, &client_info.host_info.addr,
            sizeof(struct sockaddr_storage));

    // copy data from vif_config to message
    num_vif = vif_get_names(vif_names);
    reg_info->num_vif = num_vif;
    for (i = 0; i < num_vif; i++)
        assemble_registration_data(i, &(reg_info->vif_data[i]), vif_names[i]);

    // clear wait status when timeout alarm
    signal(SIGALRM, clear_wait);

    while ((success == 0) && (message_id < 10))
    {
        reg_info->msg_id = message_id;
        wait = 1;
        debug_msg((char *) &reg_msg);

        // connect to model
        if (connect(net_socket,
                    model_info.host_info.addr_info.ai_addr,
                    model_info.host_info.addr_info.ai_addrlen) < 0)
        {
            iwc_error("Couldn't connect to model (%s)\n", strerror(errno));
            return -1;
        }

        // send register message to model
        if ((sent_bytes = sendto(net_socket, &reg_msg, REGISTRATION_MSG_SIZE, 0,
            model_info.host_info.addr_info.ai_addr,
            model_info.host_info.addr_info.ai_addrlen)) < 0)
        {
            iwc_error("Couldn't send register message (%s)\n",
                    strerror(errno));
        } else {
            iwc_debug(">>net_socket<< %i Bytes sent\n", sent_bytes);
            iwc_debug("\n");
        }

        // start waiting timer (5s) for ack message else resend register message
        alarm(5);
        while (wait == 1)
        {
            if ((recv_bytes = recv(net_socket, ack_buffer,
                ACK_MSG_SIZE, MSG_DONTWAIT)) > 0)
            {
                iwc_debug(">>net_socket<< %i Bytes read\n",
                        recv_bytes);
                if ((recv_bytes == ACK_MSG_SIZE) &&   // check for valid ack
                    is_ack_msg((struct msg_header *) ack_buffer) &&
                    (((struct plain_msg *) ack_buffer)->msg_id <= message_id))
                {
                    debug_msg(ack_buffer);
                    alarm(0);  // reset alarm
                    wait = 0;  // exit loop
                    success = 1;
                } else {
/* DEBUG            iwc_debug(">> ");
                    for (i = 0; i < recv_bytes; i++)
                        printf("%2.2x ", ack_buffer[i]);
                    printf("\n"); */
                    iwc_debug("iwc_model_register(): Discard unknown packet\n");
                }
                iwc_debug("\n");
            } else {                                // handle recv() errors
                /* When recv() is non-blocking but would block, POSIX.1-2001
                   defines the following errors (EAGAIN|EWOULDBLOCK)
                    --> in case we don't receive network reply yet! */
                if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
                {
                    if (errno == ECONNREFUSED)
                        iwc_error("Couldn't connect to WlanModel\n");
                    else
                        iwc_error("iwc_model_register(): recv() (%s)\n",
                                strerror(errno));
                    return -1;
                }
            }
        }
        message_id++;
    }
    free(ack_buffer);
    if (success == 0)
    {
        iwc_error("Couldn't register host at WlanModel\n");
        return -1;
    }
    iwc_log("Host successfully registered at WlanModel\n");

    return 1;   // return 1 to set is_registered
}

/**
 * Deregisters node from WlanModel
 * @return 0 if success, -1 else
 */
int
iwc_model_deregister(void)
{
    int error = 0;
    int sent_bytes = 0;
    struct plain_msg *dereg_msg = calloc(1, sizeof(struct plain_msg));;

    dereg_msg->header.type = DEREGISTRATION_MSG;
    strncpy(dereg_msg->header.id, client_info.host_id, 2 * ETH_ALEN + 1);
    dereg_msg->msg_id = 0;

    debug_msg((char *) dereg_msg);

    // send data to model
    if ((sent_bytes = sendto(net_socket, dereg_msg, DEREGISTRATION_MSG_SIZE, 0,
        model_info.host_info.addr_info.ai_addr,
        model_info.host_info.addr_info.ai_addrlen)) < 1)
        {
            iwc_error("iwc_model_deregister(): sendto (%s)\n", strerror(errno));
            error = -1;
        }
    else
        iwc_debug("iwc_model_deregister(): >>net_socket<< %i Bytes sent\n",
                sent_bytes);


    if (error == -1)
    {
        iwc_error("Model deregistration failed\n");
        return -1;
    }
    iwc_debug("\n");

    return 0;
}


/**
 * Sends config_msg to update wireless parameter at WlanModel
 * @param ioctl_type Ioctl type
 * @param dev_name Virtual wireless device name
 */
void
iwc_model_update(int ioctl_type, const char *dev_name)
{
    int sent_bytes = 0;

    struct config_msg *notify_msg = calloc(1, sizeof(struct config_msg));
    if (notify_msg == NULL)
        return;

    // prepare config message
    (*notify_msg).header.type = CONFIG_MSG;
    strncpy(notify_msg->header.id, client_info.host_id, 2 * ETH_ALEN + 1);

    // TODO: set vif_index!

    switch (ioctl_type)    // add new value according to ioctl
    {
        case SIOCSIWFREQ:
            // send the channel number (configure in omnetpp.ini)
            notify_msg->vif_data.value.channel = vif_get_channel(dev_name);
            break;
        case SIOCSIWSENS:
            notify_msg->vif_data.value.sensitivity =
                                                vif_get_sensitivity(dev_name);
            break;
        case SIOCSIWRETRY:
            notify_msg->vif_data.value.retry = vif_get_retry(dev_name);
            break;
        case SIOCSIWRTS:
            notify_msg->vif_data.value.rts = vif_get_rts(dev_name);
            break;
        case SIOCSIWFRAG:
            notify_msg->vif_data.value.frag = vif_get_frag(dev_name);
            break;
        case SIOCSIWTXPOW:
            notify_msg->vif_data.value.txpower = vif_get_txpower(dev_name);
            break;
        default:
            iwc_debug("iwc_model_update(): Wireless extension %x not found!\n",
                ioctl_type);
            free(notify_msg);
            return;
    }
    notify_msg->vif_data.type = ioctl_type;

    debug_msg((char *) notify_msg);

    // send data to model
    if ((sent_bytes = sendto(net_socket, notify_msg, CONFIG_MSG_SIZE, 0,
        model_info.host_info.addr_info.ai_addr,
        model_info.host_info.addr_info.ai_addrlen)) < 1)
            iwc_error("iwc_model_update(): sendto (%s)\n", strerror(errno));
    iwc_debug("iwc_model_update(): >>net_socket<< %i Bytes sent\n", sent_bytes);
    iwc_debug("\n");

    free(notify_msg);
}
